import json
import logging
import time
from functools import lru_cache

import requests

MOBILE_USER_AGENT = 'Instagram 123.0.0.21.114 (iPhone; CPU iPhone OS 11_4 like Mac OS X; en_US; en-US; scale=2.00; 750x1334) AppleWebKit/605.1.15'

logging.basicConfig()
logger = logging.getLogger('instagraph')
logger.setLevel(logging.DEBUG)


class API:
    def __init__(self,
                 user_agent=None,
                 proxy=None,
                 session=None):
        self.lang = 'it'

        if not session:
            self.session = requests.Session()
        else:
            self.session = session

        if proxy:
            proxies = {
                "http": proxy,  # Sample: "http://10.10.10.10:8000",
                "https": proxy,
            }
            self.session.proxies = proxies

        if not user_agent:
            self.user_agent = MOBILE_USER_AGENT

        self.session.headers.update({
            'User-Agent': self.user_agent})
        self.last_response = None

    @staticmethod
    def waiter(fail_count):
        fail_count += 1

        if fail_count == 10:
            raise ConnectionError(f"Connection is down. Terminating.")
        else:
            logger.warning(f"Request failed. Waiting {fail_count} seconds before retry.")
            time.sleep(fail_count)

        return fail_count

    def get(self, url):
        fail_count = 0

        logger.debug(f"Requesting {url}")

        while True:
            try:
                response = self.session.get(url)
                if response.status_code == 200:
                    logger.debug(f"Response for {url} is OK")
                    self.last_response = response.text
                    return self.shared_data
                else:
                    fail_count = self.waiter(fail_count)
            except Exception as e:
                logger.error(e)
                fail_count = self.waiter(fail_count)

    @property
    def shared_data(self):
        try:
            if self.last_response:
                shared_data = self.last_response.split('window._sharedData = ')[1].split(';</script>')[0]
                return json.loads(shared_data)
            else:
                logger.error("shared_data not found")
                return {}
        except IndexError:
            return {}

    @lru_cache(maxsize=4)
    def get_hashtag_feed(self, hashtag, max_id=None):
        url = f"https://www.instagram.com/explore/tags/{hashtag}?hl={self.lang}"
        if max_id:
            url += f"&max_id={max_id}"
        return self.get(url)

    @lru_cache(maxsize=4)
    def get_media(self, shortcode):
        url = f"https://www.instagram.com/p/{shortcode}/?hl={self.lang}"
        return self.get(url)['entry_data']['PostPage'][0]['graphql']['shortcode_media']

    @lru_cache(maxsize=4)
    def get_user(self, username):
        url = f"https://www.instagram.com/{username}/?hl={self.lang}"
        return self.get(url)['entry_data']['ProfilePage'][0]['graphql']['user']


class HashtagFeed(API):
    def __init__(self, hashtag, max_id=None, user_agent=None, proxy=None, session=None):
        super().__init__(user_agent, proxy, session)
        self.hashtag = hashtag
        shared_data = self.get_hashtag_feed(hashtag, max_id=max_id)
        self.map_shared_data(shared_data)

    def map_shared_data(self, shared_data):
        self.config = shared_data['config']
        self.country_code = shared_data['country_code']
        self.language_code = shared_data['language_code']
        self.locale = shared_data['locale']
        self.hashtag_info = shared_data['entry_data']['TagPage'][0]['graphql']['hashtag']
        self.end_cursor = self.hashtag_info['edge_hashtag_to_media']['page_info']['end_cursor']
        self.has_next_page = self.hashtag_info['edge_hashtag_to_media']['page_info']['has_next_page']
        self.count = self.hashtag_info['edge_hashtag_to_media']['count']

        self.nodes = [Node(e['node']) for e in self.hashtag_info['edge_hashtag_to_media']['edges']]

    def next_page(self):
        if self.has_next_page:
            logger.debug(f"Going to next page following end_cursor {self.end_cursor}")
            shared_data = self.get_hashtag_feed(self.hashtag, self.end_cursor)
            self.map_shared_data(shared_data)
            logger.debug(f"Done. New end_cursor is {self.end_cursor}")
            return True
        else:
            logger.warning("Feed does not have next page.")
            return False


class Node(API):
    def __init__(self, node, user_agent=None, proxy=None, session=None):
        super().__init__(user_agent, proxy, session)
        self.comments_disabled = node['comments_disabled']
        self.id = node['id']
        if node['edge_media_to_caption']['edges']:
            self.caption = node['edge_media_to_caption']['edges'][0]['node']['text']
        self.shortcode = node['shortcode']
        self.comments_count = node['edge_media_to_comment']['count']
        self.taken_at_timestamp = node['taken_at_timestamp']
        self.likes = node['edge_liked_by']['count']
        self.owner_id = node['owner']['id']
        self.thumbnail_src = node['thumbnail_src']
        self.is_video = node['is_video']
        self.accessibility_caption = node.get('accessibility_caption', None)
        self.media_info = None

    @property
    def location(self):
        self.media_info = self.get_media(self.shortcode)
        return Location(self.media_info['location'] if self.media_info['location'] else {})

    @property
    def owner_username(self):
        self.media_info = self.get_media(self.shortcode)
        return self.media_info['owner']['username']

    def location_in(self, locations):
        if self.location.name:
            for location in locations:
                loc = location.lower()
                if loc in self.location.name.lower():
                    return True
                try:
                    if loc in self.location.city_name.lower():
                        return True
                except Exception as e:
                    pass
        return False


class Location(API):
    def __init__(self, location, user_agent=None, proxy=None, session=None):
        super().__init__(user_agent, proxy, session)
        self.id = location.get('id')
        self.has_public_page = location.get('has_public_page')
        self.name = location.get('name')
        self.slug = location.get('slug')
        self.address_json = location.get('address_json')
        if self.address_json:
            self.address_json = json.loads(self.address_json)
        else:
            self.address_json = {}
        self.address = self.address_json.get('street_address')
        self.zip_code = self.address_json.get('zip_code')
        self.city_name = self.address_json.get('city_name')
        self.region_name = self.address_json.get('region_name')
        self.country_code = self.address_json.get('country_code')
        self.exact_city_match = self.address_json.get('exact_city_match')
        self.exact_region_match = self.address_json.get('exact_region_match')
        self.exact_country_match = self.address_json.get('exact_country_match')


class Profile(API):
    def __init__(self, username, user_agent=None, proxy=None, session=None):
        super().__init__(user_agent, proxy, session)
        data = self.get_user(username)
        self.data = data
        self.id = data.get('id')
        self.username = data.get('username')
        self.biography = data.get('biography')
        self.followers = data.get('edge_followed_by', {}).get('count')
        self.following = data.get('edge_follow', {}).get('count')
        self.full_name = data.get('full_name')
        self.is_business_account = data.get('is_business_account')
        self.is_joined_recently = data.get('is_joined_recently')
        self.is_private = data.get('is_private')
        self.is_verified = data.get('is_verified')
        self.profile_pic_url = data.get('profile_pic_url')
        _nodes = data.get('edge_owner_to_timeline_media', {})
        self.medias_count = _nodes.get('count')
        self.has_next_page = _nodes.get('page_info', {}).get('has_next_page', False)
        self.end_cursor = _nodes.get('page_info', {}).get('end_cursor')
        self.nodes = [Node(e['node']) for e in _nodes['edges']]

    def followers_between(self, min, max):
        if min <= self.followers <= max:
            return True
        return False

    def check_keyword(self, keywords, elements=10):
        count = 0
        for media in self.nodes[:elements]:
            if media.accessibility_caption:
                for keyword in keywords:
                    if keyword in media.accessibility_caption:
                        count += 1
        return count

    def check_locations(self, locations):
        if locations:
            for location in locations:
                if location.lower() in self.biography.lower():
                    return True
        return False
